<?php

namespace Drupal\Tests\message_digest\Kernel;

use Drupal\message\Entity\Message;

/**
 * Kernel tests for Message Digest.
 *
 * @group message_digest
 */
class MessageDigestTest extends DigestTestBase {

  /**
   * Tests the plugin deriver for daily and weekly digests.
   */
  public function testDigestNotifierPluginsExist() {
    $count = 0;
    foreach ($this->notifierManager->getDefinitions() as $plugin_id => $plugin_definition) {
      if ($plugin_definition['provider'] === 'message_digest') {
        $dummy = Message::create(['template' => 'foo']);
        // Ensure the plugin can be instantiated.
        $this->notifierManager->createInstance($plugin_id, [], $dummy);
        $count++;
      }
    }
    $this->assertEquals(2, $count, 'There are 2 digest notifiers.');
  }

  /**
   * Tests the notifier sending and delivery.
   */
  public function testNotifierDelivery() {
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $dummy = Message::create(['template' => $template->id()]);
    /** @var \Drupal\message_digest\Plugin\Notifier\DigestInterface $digest_notifier */
    $digest_notifier = $this->notifierManager->createInstance('message_digest:daily', [], $dummy);

    $account = $this->createUser();
    $dummy->setOwner($account);
    $dummy->save();
    $this->notifierSender->send($dummy, [], $digest_notifier->getPluginId());
    $result = $this->container->get('database')
      ->select('message_digest', 'm')
      ->fields('m')
      ->execute()
      ->fetchAllAssoc('id');
    $this->assertEquals(1, count($result));
    foreach ($result as $row) {
      $this->assertEquals($account->id(), $row->receiver);
      $this->assertEquals($digest_notifier->getPluginId(), $row->notifier);
    }

    // Now deliver the message.
    $this->container->get('message_digest.manager')->processDigests();
    $this->assertMail('subject', 'Drupal message digest', 'Expected email subject is set.');
    // @todo Fix the rendering of the message body.
    // @see https://www.drupal.org/node/2813383
    // $this->assertMail('body', "foo bar\n", 'Expected email body is set.');
    $this->assertMail('id', 'message_digest_digest', 'Expected email key is set.');
    $this->assertMail('to', $account->getEmail(), 'Expected email recipient is set.');

    // Verify that hook_message_digest_view_mode_alter() has been called.
    // @see message_digest_test_message_digest_view_mode_alter().
    $this->assertTrue($this->container->get('state')->get('message_digest_test_message_digest_view_mode_alter', FALSE));
  }

  /**
   * Tests message aggregation.
   */
  public function testNotifierAggregation() {
    // Send several messages.
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $account = $this->createUser();
    $message_1 = Message::create(['template' => $template->id()]);
    $message_2 = Message::create(['template' => $template->id()]);
    /** @var \Drupal\message_digest\Plugin\Notifier\DigestInterface $digest_notifier */
    $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', [], $message_1);

    $message_1->setOwner($account);
    $message_1->save();
    $message_2->setOwner($account);
    $message_2->save();
    $this->notifierSender->send($message_1, [], $digest_notifier->getPluginId());
    $this->notifierSender->send($message_2, [], $digest_notifier->getPluginId());
    $result = $this->container->get('database')
      ->select('message_digest', 'm')
      ->fields('m')
      ->execute()
      ->fetchAllAssoc('id');
    $this->assertEquals(2, count($result));
    foreach ($result as $row) {
      $this->assertEquals($account->id(), $row->receiver);
      $this->assertEquals($digest_notifier->getPluginId(), $row->notifier);
    }

    // Aggregate and mark processed.
    $results = $digest_notifier->aggregate();
    $digest_notifier->setLastSent();
    $expected = [
      $account->id() => [
        '' => [
          '' => [
            $message_1->id(),
            $message_2->id(),
          ],
        ],
      ],
    ];
    $this->assertSame($expected, $results);

    // Verify that the aggregate alter hook was called.
    // @see message_digest_test_message_digest_aggregate_alter()
    $this->assertTrue($this->container->get('state')->get('message_digest_test_message_digest_aggregate_alter', FALSE));

    // Since this has been marked as sent, calling aggregate again should result
    // in an empty array.
    $this->assertEmpty($digest_notifier->aggregate());

    // Set last sent time to 8 days in the past.
    $last_run = strtotime('-8 days', $this->container->get('request_stack')->getCurrentRequest()->server->get('REQUEST_TIME'));
    $this->container->get('state')->set($digest_notifier->getPluginId() . '_last_run', $last_run);
    $results = $digest_notifier->aggregate();
    $this->assertSame($expected, $results);

    // Aggregate should not return any results once marked sent.
    $digest_notifier->markSent($account);
    $this->assertEmpty($digest_notifier->aggregate());
  }

  /**
   * Test grouping by entity type, and ID.
   */
  public function testDigestGrouping() {
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $account = $this->createUser();

    // Send several messages w/o grouping.
    $global_messages = [];
    foreach (range(1, 3) as $i) {
      $message = Message::create(['template' => $template->id()]);
      $message->setOwner($account);
      $message->save();
      $global_messages[$i] = $message;
      $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', [], $message);
      $this->notifierSender->send($message, [], $digest_notifier->getPluginId());
    }

    // Now send some grouped by entity type.
    $configuration = [
      'entity_type' => 'foo',
      'entity_id' => 'bar',
    ];
    $grouped_messages = [];
    foreach (range(1, 3) as $i) {
      $message = Message::create(['template' => $template->id()]);
      $message->setOwner($account);
      $message->save();
      $grouped_messages[$i] = $message;
      $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', $configuration, $message);
      $this->notifierSender->send($message, $configuration, $digest_notifier->getPluginId());
    }

    // Aggregate and mark processed.
    $results = $digest_notifier->aggregate();
    $digest_notifier->setLastSent();
    $expected = [
      $account->id() => [
        '' => [
          '' => ['1', '2', '3'],
        ],
        'foo' => [
          'bar' => ['4', '5', '6'],
        ],
      ],
    ];
    $this->assertSame($expected, $results);
  }

  /**
   * Tests sending with an entity_id and no type.
   *
   * @expectedException \Drupal\message_digest\Exception\InvalidDigestGroupingException
   * @expectedExceptionMessage Tried to create a message digest without both entity_type () and entity_id (42). These either both need to be empty, or have values.
   */
  public function testInvalidEntityType() {
    $configuration = [
      'entity_id' => 42,
    ];
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $message = Message::create(['template' => $template->id()]);
    $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', $configuration, $message);
    $this->notifierSender->send($message, $configuration, $digest_notifier->getPluginId());
  }

  /**
   * Tests sending with an entity_type and no ID.
   *
   * @expectedException \Drupal\message_digest\Exception\InvalidDigestGroupingException
   * @expectedExceptionMessage Tried to create a message digest without both entity_type (foo) and entity_id (). These either both need to be empty, or have values.
   */
  public function testInvalidEntityId() {
    $configuration = [
      'entity_type' => 'foo',
    ];
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $message = Message::create(['template' => $template->id()]);
    $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', $configuration, $message);
    $this->notifierSender->send($message, $configuration, $digest_notifier->getPluginId());
  }

  /**
   * Test old message purging.
   */
  public function testMessageCleanup() {
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', []);
    $account = $this->createUser();
    $messages = [];
    // Send 10 messages.
    foreach (range(1, 10) as $i) {
      $message = Message::create(['template' => $template->id()]);
      $message->setOwner($account);
      $message->save();
      $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', [], $message);
      $this->notifierSender->send($message, [], $digest_notifier->getPluginId());
      $messages[$i] = $message;
    }
    $digest = $digest_notifier->aggregate();
    $this->assertEquals(10, count($digest[$account->id()]['']['']));
    $digest_notifier->markSent($account);

    // Delete 5 messages.
    foreach (range(1, 5) as $i) {
      $messages[$i]->delete();
    }
    $this->container->get('message_digest.manager')->cleanupOldMessages();
    $result = $this->container->get('database')->select('message_digest', 'md')
      ->fields('md')
      ->execute()
      ->fetchAllAssoc('id');
    $this->assertEquals(5, count($result));
    foreach ($result as $row) {
      $this->assertGreaterThan(5, $row->mid);
    }
  }

  /**
   * Tests the message digest manager processing.
   */
  public function testManagerProcessing() {
    $this->container->get('message_digest.manager')->processDigests();
    $request_time = $this->container->get('request_stack')->getCurrentRequest()->server->get('REQUEST_TIME');
    $this->assertEquals($request_time, $this->container->get('state')->get('message_digest:daily_last_run'));
    $this->assertEquals($request_time, $this->container->get('state')->get('message_digest:weekly_last_run'));
    $this->assertEmpty($this->getMails());

    // Actually add some messages now, and reset last sent time.
    $last_run = strtotime('-8 days', $this->container->get('request_stack')->getCurrentRequest()->server->get('REQUEST_TIME'));
    $this->container->get('state')->set('message_digest:weekly_last_run', $last_run);
    $template = $this->createMessageTemplate('foo', 'Foo', 'Foo, foo', ['<p>foo</p>', '<h2>bar</h2>']);
    $account = $this->createUser();
    $messages = [];
    // Send 10 messages.
    foreach (range(1, 10) as $i) {
      $message = Message::create(['template' => $template->id()]);
      $message->setOwner($account);
      $message->save();
      $digest_notifier = $this->notifierManager->createInstance('message_digest:weekly', [], $message);
      $this->notifierSender->send($message, [], $digest_notifier->getPluginId());
      $messages[$i] = $message;
    }
    // Ensure cron function actually calls the process method.
    $this->container->get('cron')->run();
    $this->assertEquals(1, count($this->getMails()));
    $this->assertMail('to', $account->getEmail());
  }

}
