<?php

namespace Drupal\Tests\message_digest\Unit\Plugin\Notifier;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Insert;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\State\StateInterface;
use Drupal\message\MessageInterface;
use Drupal\message_digest\Plugin\Notifier\Digest;
use Drupal\Tests\UnitTestCase;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Tests the Digest plugin.
 *
 * @group message_digest
 *
 * @coversDefaultClass \Drupal\message_digest\Plugin\Notifier\Digest
 */
class DigestTest extends UnitTestCase {

  /**
   * Digest configuration.
   *
   * @var array
   */
  protected $configuration = [];

  /**
   * Mocked database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * Mocked entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Plugin definition.
   *
   * @var array
   */
  protected $pluginDefinition = [
    'digest_interval' => '1 day',
  ];

  /**
   * Plugin ID.
   *
   * @var string
   */
  protected $pluginId;

  /**
   * Mocked request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Mocked state.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * {@inheritdoc}
   */
  public function setUp() {
    parent::setUp();

    $this->connection = $this->prophesize(Connection::class)->reveal();
    $this->entityTypeManager = $this->prophesize(EntityTypeManagerInterface::class)->reveal();
    $this->pluginId = $this->randomMachineName();
    $this->requestStack = $this->prophesize(RequestStack::class)->reveal();
    $this->state = $this->prophesize(StateInterface::class)->reveal();

  }

  /**
   * Test delivery without a saved message entity.
   *
   * @covers ::deliver
   *
   * @expectedException \AssertionError
   * @expectedExceptionMessage The message entity (or $message->original) must be saved in order to create a digest entry.
   */
  public function testDeliverUnsavedMessage() {
    // Setup an unsaved message.
    $message = $this->prophesize(MessageInterface::class)->reveal();
    $notifier = $this->getNotifier($message);
    $notifier->deliver([]);
  }

  /**
   * Test delivery without a saved original message entity.
   *
   * @covers ::deliver
   *
   * @expectedException \AssertionError
   * @expectedExceptionMessage The message entity (or $message->original) must be saved in order to create a digest entry.
   */
  public function testDeliveryUnsavedOriginal() {
    // Setup an unsaved original message.
    $original = $this->prophesize(MessageInterface::class)->reveal();
    $message = $this->prophesize(MessageInterface::class)->reveal();
    $message->original = $original;

    $notifier = $this->getNotifier($message);
    $notifier->deliver([]);
  }

  /**
   * Test delivery with an unsaved message, but a saved original.
   *
   * @covers ::deliver
   */
  public function testDeliverySavedOriginal() {
    // Setup a saved original.
    $original = $this->prophesize(MessageInterface::class);
    $original->id()->willReturn(42);
    $message = $this->prophesize(MessageInterface::class);
    $message->getOwnerId()->willReturn(4);
    $message->getCreatedTime()->willReturn(123);
    $message->id()->willReturn(NULL);
    $message = $message->reveal();
    $message->original = $original->reveal();

    // Mock up the insert.
    $expected_row = [
      'receiver' => 4,
      'entity_type' => '',
      'entity_id' => '',
      'notifier' => $this->pluginId,
      'timestamp' => 123,
      'mid' => 42,
    ];
    $insert = $this->prophesize(Insert::class);
    $insert->fields($expected_row)->willReturn($insert->reveal());
    $insert->execute()->shouldBeCalled();
    $connection = $this->prophesize(Connection::class);
    $connection->insert('message_digest')->willReturn($insert->reveal());
    $this->connection = $connection->reveal();

    $notifier = $this->getNotifier($message);
    $this->assertTrue($notifier->deliver([]));
  }

  /**
   * Helper method to construct a digest notifier.
   *
   * @param \Drupal\message\MessageInterface $message
   *   Mocked message for the notifier.
   *
   * @return \Drupal\message_digest\Plugin\Notifier\Digest
   *   The message notifier.
   */
  protected function getNotifier(MessageInterface $message) {
    $logger = $this->prophesize(LoggerChannelInterface::class)->reveal();

    $notifier = new Digest(
      $this->configuration,
      $this->pluginId,
      $this->pluginDefinition,
      $logger,
      $this->entityTypeManager,
      $message,
      $this->state,
      $this->connection,
      $this->requestStack
    );

    return $notifier;
  }

}
